#!/bin/bash


# TO DO:
#
# The audio head/tail parts are re-sampled seperately from
# the body parts, which causes discontinuities in the sound.
# We should extract the entire soundtrack, then resample it
# to 48 KHz, then cut the head and tail.
#
#
# Audio discontinuities can occur when

SMIL2WavHeaderSize=40

# ----------------------------------------------------------
# COMMUNICATION ROUTINES FOR USE WITH KOMMANDER GUI
# ----------------------------------------------------------

# Set the currently executing task to $1.

UpdateTask ()
{
 if [ -n "$DialogDCOPTaskCommand" ]; then
  $DialogDCOPTaskCommand "$1"
 fi
}

# Update progress to $1 %.

UpdateProgress ()
{
 if [ -n "$DialogDCOPProgressCommand" ]; then
   $DialogDCOPProgressCommand $1
 fi
}


# ----------------------------------------------------------
# ROUTINES TO COMMUNICATE ERRORS/WARNINGS TO THE USER
# ----------------------------------------------------------


# Present an error message to the user
# Argument $1 is error message.

PrintError ()
{
 X=`type -p kdialog`

 if test $X && test -x $X && [ -n "$DISPLAY" ]; then
  kdialog --error "$1"
 else
  echo
  echo "---- RECEIVED ERROR MESSAGE ----"
  echo
  echo -e "$1"
  echo
  echo "--------------------------------"
 fi
 exit -1
}

# Present a warning message to the user
# Argument $1 is warning message.

PrintWarning ()
{
 X=`type -p kdialog`

 if test $X && test -x $X && [ -n "$DISPLAY" ]; then
  kdialog --warningcontinuecancel "$1"
  if [ ! "$?" -eq 0 ]; then
  echo
  echo "USER ABORTED PROGRAM"
  echo
  exit -1
  fi
  
 else
  echo
  echo "---- WARNING ----"
  echo
  echo -e "$1"
  echo
  echo "-----------------"
 fi
}

# ----------------------------------------------------------
# CHECKS TO SEE IF WE HAVE ALL THE REQUIRED TOOLS AVAILABLE.
# ----------------------------------------------------------

Requirements=( convert composite sox smil2wav smil2raw kdialog )

PrettyRequirements=( "ImageMagick >= 6" "Sox" "SMIL Utils" "KDialog" )

# Check if tool $1 is available.

ToolCheck ()
{
 echo -ne "Checking for \"$1\" ...\t"
	X=`type -p $1`
	if test $X && test -x $X;
		then echo " found." ; return 0
		else echo " MISSING" ; return 1
	fi
}

# Print an error when a required utility is missing

Missing ()
{
 PrintError "You don't have the utility \"$1\" installed! \n\nTo use this progam, you need to have at least these packages installed:\n\n`
 
 Requirement=0
 while [ -n "${PrettyRequirements[$Requirement]}" ]
 do
  echo -n "* ${PrettyRequirements[$Requirement]}\n"
  Requirement=$[$Requirement + 1]
 done
 
 
 `" 
 exit -1
}

RunChecks ()
{
 Requirement=0
 while [ -n "${Requirements[$Requirement]}" ]
 do
   ToolCheck "${Requirements[$Requirement]}" || Missing "${Requirements[$Requirement]}"
   Requirement=$[$Requirement + 1]
 done
 echo
}


# ----------------------------------------------------------
# DISK SPACE CHECKING ROUTINES
# ----------------------------------------------------------

# Warn if disk space in directory $1 lower than $2 MB

DiskSpaceWarning ()
{
  FreeSpace=`df  -m "$1" -P | awk '{print $4}' | tail -n 1`

  if [ $FreeSpace -lt $2 ]; then
  PrintWarning "You are low on disk space in the following directory:\n\n$1\n\nTo be completely safe, you are advised to free more disk space before continuing."
  fi
}

# Abort if disk space ran out in directory $1

DiskSpaceError ()
{
  FreeSpace=`df  -m "$1" -P | awk '{print $4}' | tail -n 1`
  
  if [ $FreeSpace -le 1 ]; then
  PrintError "This program ran out of disk space in the following directory:\n\n$1\n\nPlease free some more disk space and try again."
  fi
}

# Probe resolution of video $1

DetectVideoRes ()
{
 ffmpeg -i "$1" -f yuv4mpegpipe -pix_fmt yuv420p -t 1 -y /tmp/tmp.y4m >/dev/null 2>&1

 VerticalVideoRes=`head -n 1 /tmp/tmp.y4m | awk '{sub("W","",$2); sub("H","",$3); print $3}'`

 rm -f /tmp/tmp.y4m
}

# Probe type of video: PAL or NTSC.

DetectVideoType ()
{
 if [ "$VerticalVideoRes" == 576 ]; then

   echo "PAL"

 else

   echo "NTSC"

 fi
}

# Probe aspect ratio (widescreen or standard) of video $1

DetectVideoAspect ()
{
 if [ "$(tcprobe -i "$1" | grep 'aspect ratio' | grep -E -o [0-9]{1,\}:[0-9]{1,\})" == "16:9" ]; then

  echo "wide"

 else

  echo "standard"

 fi
}


# Translate a frame index $1 to a time index

TimeIndex ()
{
 Hours=$( echo "scale=2; ($1 / $FrameRate) / 3600" | bc | cut -d . -f 1)

 if [ -z "$Hours" ]; then
  Hours=0
 fi

 Minutes=$( echo "scale=2; (($1 / $FrameRate) - ($Hours * 60) ) / 60" | bc | cut -d . -f 1)

 if [ -z "$Minutes" ]; then
  Minutes=0
 fi

 Seconds=$( echo "scale=2; ($1 / $FrameRate) - ($Hours * 3600) - ($Minutes * 60)" | bc | cut -d . -f 1)

 if [ -z "$Seconds" ]; then
  Seconds=0
 fi

 echo "$Hours:$Minutes:$Seconds"
}


# Takes DV file $1 and saves first $2 milliseconds to DV file $3

DVHead ()
{
 if [ $2 -gt 0 ]; then

  UpdateTask "Extracting head of video file..."
 
  echo "<?xml version=\"1.0\"?>" > "$WorkingDir/Head.smil"
  echo "<smil xmlns:smil2=\"http://www.w3.org/2001/SMIL20/Language\">" >> "$WorkingDir/Head.smil"
  echo "  <seq>" >> "$WorkingDir/Head.smil"
  echo "    <video src=\"$1\" clipBegin=\"0\" clipEnd=\"$(( ($2 / $SingleFrameDuration) - 1 ))\"/>" >> "$WorkingDir/Head.smil"
  echo "  </seq>" >> "$WorkingDir/Head.smil"
  echo "</smil> " >> "$WorkingDir/Head.smil"

  smil2raw "$WorkingDir/Head.smil" > "$3"
  
  DiskSpaceError "$3"
  
  rm -f "$WorkingDir/Head.smil"
  
 fi
}

# Takes DV file $1 and saves last $2 milliseconds to DV file $3

DVTail ()
{
 if [ $2 -gt 0 ]; then
  
  UpdateTask "Extracting tail of video file..."
  
  
  NumTailFrames=$[ `du -b "$1" | awk '{print $1}'` / $DVFrameSize ]
  
  echo "<?xml version=\"1.0\"?>" > "$WorkingDir/Tail.smil"
  echo "<smil xmlns:smil2=\"http://www.w3.org/2001/SMIL20/Language\">" >> "$WorkingDir/Tail.smil"
  echo "  <seq>" >> "$WorkingDir/Tail.smil"
  echo "    <video src=\"$1\" clipBegin=\"$(($NumTailFrames - ($2 / $SingleFrameDuration) ))\" clipEnd=\"$(( $NumTailFrames - 1 ))\"/>" >> "$WorkingDir/Tail.smil"
  echo "  </seq>" >> "$WorkingDir/Tail.smil"
  echo "</smil> " >> "$WorkingDir/Tail.smil"

  smil2raw "$WorkingDir/Tail.smil" > "$3"
  
  DiskSpaceError "$3"
  rm -f "$WorkingDir/Tail.smil"
  
 fi
}


# Detect the properties of the soundtrack of
# DV video $1
# Returns properties in global variables
# DetectedSampleRate and DetectedBytesPerSecond

DetectSoundProperties ()
{
  # Extract the header from the soundtrack of the DV video
  smil2wav "$1" 2>/dev/null | dd bs=$SMIL2WavHeaderSize count=1 of="$WorkingDir/header" 
  
  # Examine header, find sample rate
  DetectedSampleRate=`cat "$WorkingDir/header" | sox -V -t .wav /dev/stdin -e 2>&1 | grep -E -o [0-9]{1,\}\ samp/sec | awk '{print $1}'`
  
  if [ -z "$DetectedSampleRate" ]; then
    PrintError "Something is wrong with the soundtrack of video file $1.\n\nFailed to detect the correct sample rate."
  elif [ "$DetectedSampleRate" == "0" ]; then
    PrintError "The soundtrack of video file $1 is damaged. It cannot be used.\n\nYou can try to fix this by removing the first couple of frames from the video using Kino or similar editing software."
  else
    echo "Detected sample rate of $DetectedSampleRate"
  fi
  
  # Examine header, find bytes per second
  DetectedBytesPerSecond=`cat "$WorkingDir/header" | sox -V -t .wav /dev/stdin -e 2>&1 | grep -E -o [0-9]{1,\}\ byte/sec | awk '{print $1}'`
  
  if [ -z "$DetectedBytesPerSecond" ]; then
    PrintError "Something is wrong with the soundtrack of video file $1.\n\nFailed to detect the correct bytes-per-second value."
    exit -1
  else
    echo "Detected Bytes-Per-Second value of $DetectedBytesPerSecond"
  fi
  
  rm -f "$WorkingDir/header"
}


# Extracts the head from the soundtrack of a DV file.
#
# $1 = DV Video file
# $2 = Length of head in milliseconds
# $3 = Output PCM Wave file

SoundHead ()
{
 if [ $2 -gt 0 ]; then
  
  UpdateTask "Extracting head of soundtrack..."
  
  echo "Extracting $2 milliseconds of head sound from $1"
  
  DetectSoundProperties "$1"
 
  if [ ! "$DetectedSampleRate" == "48000" ]; then
    
    smil2wav "$1" 2>/dev/null | dd bs=$SMIL2WavHeaderSize skip=1 | head -c $(( ($DetectedBytesPerSecond * $2) / 1000 )) | sox -r $DetectedSampleRate -s -c 2 -w -t .raw /dev/stdin -r 48000 "$3" resample
  
  else
  
    smil2wav "$1" 2>/dev/null | dd bs=$SMIL2WavHeaderSize skip=1 | head -c $(( ($DetectedBytesPerSecond * $2) / 1000 )) | sox -r $DetectedSampleRate -s -c 2 -w -t .raw /dev/stdin "$3"
  
  fi
 
  DiskSpaceError "$3"
 
 fi 
}

# Extracts the tail from the soundtrack of a DV file.
#
# $1 = DV Video file
# $2 = Length of tail in milliseconds
# $3 = Output PCM Wave file

SoundTail ()
{
 if [ $2 -gt 0 ]; then
  
  UpdateTask "Extracting tail of soundtrack..."
  
  echo "Extracting $2 milliseconds of tail sound from $1"
  
  DetectSoundProperties "$1"
  
  if [ ! $DetectedSampleRate == 48000 ]; then
  
    smil2wav "$1" 2>/dev/null | dd bs=$SMIL2WavHeaderSize skip=1 | tail -c $(( ($DetectedBytesPerSecond * $2) / 1000 )) | sox -r $DetectedSampleRate -s -c 2 -w -t .raw /dev/stdin -r 48000 "$3" resample
 
  else
  
    smil2wav "$1" 2>/dev/null | dd bs=$SMIL2WavHeaderSize skip=1 | tail -c $(( ($DetectedBytesPerSecond * $2) / 1000 )) | sox -r $DetectedSampleRate -s -c 2 -w -t .raw /dev/stdin "$3"
 
  fi
 
  DiskSpaceError "$3"
 
 fi
}


# Video Transition Effect 'Crossfade'
#
# $1 = DV Tail filename
# $2 = DV Head filename
# $3 = Length of Head and Tail parts, in milliseconds
# $4 = Output filename

VideoTransEffect_Crossfade ()

{
  echo -n "Generating Crossfade frames"

  rm -f -R "$WorkingDir/HeadFrames" "$WorkingDir/TailFrames" "$WorkingDir/TransitionFrames"
  
  mkdir "$WorkingDir/HeadFrames"
  mkdir "$WorkingDir/TailFrames"
  mkdir "$WorkingDir/TransitionFrames"

  NumTransFrames=$[$3 / $SingleFrameDuration]
  
  UpdateTask "Extracting frames from video..."
  
  pushd "$WorkingDir"
   
  DiskSpaceWarning "$WorkingDir" $NumTransFrames
  
  cat "$1" | raw2image -i 0 -m TailFrames/.png
  cat "$2" | raw2image -i 0 -m HeadFrames/.png
  
  DiskSpaceError "$WorkingDir"
  
  popd
  
  DiskSpaceWarning "$WorkingDir/TransitionFrames" $NumTransFrames
  
  Blend=0
  FrameCounter=1
  
  UpdateTask "Creating crossfade effect..."
  UpdateProgress 0

  for Frame in `ls "$WorkingDir/HeadFrames"`
    do
    
    composite -blend $Blend "$WorkingDir/HeadFrames/$Frame" "$WorkingDir/TailFrames/$Frame" +matte -depth 8 "$WorkingDir/TransitionFrames/$Frame"
    
    Blend=`echo "scale=2; ($FrameCounter * 100 )/ $NumTransFrames" | bc`
    FrameCounter=$[$FrameCounter + 1]
    
    echo -n "."

    UpdateProgress $[ ($FrameCounter * 100) / $NumTransFrames ]
       
    done
  
  DiskSpaceError "$WorkingDir"
  
  UpdateProgress 100  
    
  echo "Encoding to DV Video..."

  UpdateTask "Encoding to MiniDV Video..."
    
  pushd "$WorkingDir/TransitionFrames/"
  
  image2raw * > "$4"
    
  popd

  DiskSpaceError "$WorkingDir/TransitionFrames/"
   
  echo "Done".
  
  UpdateTask "Done."
}

# Video Transition Effect 'PushAside'
#
# $1 = DV Tail filename
# $2 = DV Head filename
# $3 = Length of Head and Tail parts, in milliseconds
# $4 = Output filename

VideoTransEffect_PushAside ()

{
  echo -n "Generating PushAside frames"

  rm -f -R "$WorkingDir/HeadFrames" "$WorkingDir/TailFrames" "$WorkingDir/TransitionFrames"
  
  mkdir "$WorkingDir/HeadFrames"
  mkdir "$WorkingDir/TailFrames"
  mkdir "$WorkingDir/TransitionFrames"

  UpdateTask "Extracting frames from video file..."
  
  NumTransFrames=$[$3 / $SingleFrameDuration]
  
  pushd "$WorkingDir"
   
  DiskSpaceWarning "$WorkingDir" $[ $NumTransFrames ]
  
  cat "$1" | raw2image -i 0 -m TailFrames/.png
  cat "$2" | raw2image -i 0 -m HeadFrames/.png
  
  popd

  DiskSpaceWarning "$WorkingDir/TransitionFrames" $[ $NumTransFrames ]

  FrameCounter=1
  echo "processing $NumTransFrames frames:"
  UpdateTask "Creating push-aside effect..."
  UpdateProgress 0

  for Frame in `ls "$WorkingDir/HeadFrames"`
    do

    Push=$[ ($FrameCounter * 720 ) / $NumTransFrames ]

    if [ $Push == 0 ]; then

      convert "$WorkingDir/TailFrames/$Frame" +matte -depth 8 "$WorkingDir/TransitionFrames/$Frame"      

    elif [ $Push == 720 ]; then  

      convert "$WorkingDir/HeadFrames/$Frame" +matte -depth 8 "$WorkingDir/TransitionFrames/$Frame"      

    else 

      convert "$WorkingDir/HeadFrames/$Frame" -crop ${Push}x$VerticalVideoRes+0+0 "$WorkingDir/PartA.png"
      convert "$WorkingDir/TailFrames/$Frame" -crop $[ 720 - $Push]x$VerticalVideoRes+${Push}+0 "$WorkingDir/PartB.png"
      montage "$WorkingDir/PartB.png" "$WorkingDir/PartA.png" -mode Concatenate -tile x1 +matte -depth 8 "$WorkingDir/TransitionFrames/$Frame"

    fi


    FrameCounter=$[$FrameCounter + 1]

    echo -n "."
    UpdateProgress $[ ($FrameCounter * 100) / $NumTransFrames ]

    done

  DiskSpaceError "$WorkingDir"

  UpdateProgress 100  
  UpdateTask "Encoding to MiniDV Video..."

  pushd "$WorkingDir/TransitionFrames/"

  image2raw * > "$4"

  popd

  UpdateTask "Done."

  DiskSpaceError "$WorkingDir/TransitionFrames/"

  echo "Done".
}


# Video Transition Effect 'Flash'
#
# $1 = DV Tail filename
# $2 = DV Head filename
# $3 = Length of Head and Tail parts, in milliseconds
# $4 = Output filename

VideoTransEffect_Flash ()

{
  echo -n "Generating Flash frames"

  rm -f -R "$WorkingDir/HeadFrames" "$WorkingDir/TailFrames" "$WorkingDir/TransitionFrames"

  mkdir "$WorkingDir/HeadFrames"
  mkdir "$WorkingDir/TailFrames"
  mkdir "$WorkingDir/TransitionFrames"

  UpdateTask "Extracting frames from video file..."

  NumTransFrames=$[$3 / $SingleFrameDuration]

  pushd "$WorkingDir"

  DiskSpaceWarning "$WorkingDir" $[ $NumTransFrames ]

  cat "$1" | raw2image -i 0 -m TailFrames/.png
  cat "$2" | raw2image -i 0 -m HeadFrames/.png

  popd

  DiskSpaceWarning "$WorkingDir/TransitionFrames/$Frame" $[ $NumTransFrames ]

  Blend=0
  FrameCounter=1

  convert -size 720x$VerticalVideoRes xc:white "$WorkingDir/White.png"
  UpdateTask "Creating Flash effect..."
  UpdateProgress 0

  for Frame in `ls "$WorkingDir/TailFrames"`
    do
    NextFrameNumber=`echo $[1000000 + $FrameCounter] | tail -c 7`

    composite -blend $Blend "$WorkingDir/White.png" "$WorkingDir/TailFrames/$Frame" +matte -depth 8 "$WorkingDir/TransitionFrames/$NextFrameNumber.png"

    Blend=`echo "scale=2; ($FrameCounter * 100) / $NumTransFrames" | bc`
    FrameCounter=$[$FrameCounter + 1]

    echo -n "."
    UpdateProgress $[ ($FrameCounter * 100) / ($NumTransFrames * 2) ]

    done

  DiskSpaceError "$WorkingDir"

  for Frame in `ls "$WorkingDir/HeadFrames"`
    do

    Blend=`echo "scale=2; 100 - (($FrameCounter - $NumTransFrames) * 100) / $NumTransFrames" | bc`

    NextFrameNumber=`echo $[1000000 + $FrameCounter] | tail -c 7`

    composite -blend $Blend "$WorkingDir/White.png" "$WorkingDir/HeadFrames/$Frame" +matte -depth 8 "$WorkingDir/TransitionFrames/$NextFrameNumber.png"

    FrameCounter=$[$FrameCounter + 1]

    echo -n "."
    UpdateProgress $[ ($FrameCounter * 100) / ($NumTransFrames * 2) ]

    done

  DiskSpaceError "$WorkingDir"

  UpdateProgress 100  
  UpdateTask "Encoding to MiniDV Video..."

  pushd "$WorkingDir/TransitionFrames/"

  image2raw * > "$4"

  popd

  UpdateTask "Done."

  DiskSpaceError "$WorkingDir/TransitionFrames/"

  echo "Done".
}

# Video Transition Effect 'Blur'
#
# $1 = DV Tail filename
# $2 = DV Head filename
# $3 = Length of Head and Tail parts, in milliseconds
# $4 = Output filename

VideoTransEffect_Blur ()

{
  echo -n "Generating Blur frames"

  rm -f -R "$WorkingDir/HeadFrames" "$WorkingDir/TailFrames" "$WorkingDir/TransitionFrames"

  mkdir "$WorkingDir/HeadFrames"
  mkdir "$WorkingDir/TailFrames"
  mkdir "$WorkingDir/TransitionFrames"

  NumTransFrames=$[($3 / $SingleFrameDuration) * 2]

  UpdateTask "Extracting frames from video file..."

  pushd "$WorkingDir"

  DiskSpaceWarning "$WorkingDir" $[ $NumTransFrames ]

  cat "$1" | raw2image -i 0 -m TailFrames/.png
  cat "$2" | raw2image -i 0 -m HeadFrames/.png

  popd

  DiskSpaceWarning "$WorkingDir/TransitionFrames/$Frame" $[ $NumTransFrames ]

  NumFramesLeft=$NumTransFrames

  FrameCounter=1

  convert -size 720x$VerticalVideoRes xc:white "$WorkingDir/White.png"

  FrameNames=( `ls "$WorkingDir/TailFrames"` )
  NumFrames=$(( $NumFramesLeft / 2 ))
  UpdateTask "Creating Blur effect..."
  UpdateProgress 0

  for ((Frame=0; Frame < $NumFrames ; Frame++))
    do

    Resolution=$(( 1000 - (($Frame * 1000) / ($NumFrames + 1) ) ))

    ReducedXRes=$(( (720 * $Resolution) / 1000 ))
    ReducedYRes=$(( ($VerticalVideoRes * $Resolution) / 1000 ))

    if [ $ReducedXRes -lt 2 ]; then
     ReducedXRes=2
    fi
    if [ $ReducedYRes -lt 2 ]; then
     ReducedYRes=2
    fi

    NextFrameNumber=`echo $[1000000 + $FrameCounter] | tail -c 7`

    convert "$WorkingDir/TailFrames/${FrameNames[$Frame]}" -resize ${ReducedXRes}x${ReducedYRes}! -blur 0.2 -resize 720x$VerticalVideoRes! +matte -depth 8 "$WorkingDir/TransitionFrames/$NextFrameNumber.png"

    FrameCounter=$[$FrameCounter + 1]
    NumFramesLeft=$[ $NumFramesLeft - 1 ]

    echo -n "."
    UpdateProgress $[ ($FrameCounter * 100) / ($NumTransFrames * 2) ]

    done

  DiskSpaceError "$WorkingDir"

  FrameNames=( `ls "$WorkingDir/HeadFrames"` )
  NumFrames=$NumFramesLeft

  for ((Frame=0; Frame < $NumFrames ; Frame++))
    do
    Resolution=$(( (($Frame * 1000) / $NumFrames ) ))

    ReducedXRes=$(( (720 * $Resolution) / 1000 ))
    ReducedYRes=$(( ($VerticalVideoRes * $Resolution) / 1000 ))

    if [ $ReducedXRes -lt 2 ]; then
     ReducedXRes=2
    fi
    if [ $ReducedYRes -lt 2 ]; then
     ReducedYRes=2
    fi

    NextFrameNumber=`echo $[1000000 + $FrameCounter] | tail -c 7`

    convert "$WorkingDir/HeadFrames/${FrameNames[$Frame]}" -resize ${ReducedXRes}x${ReducedYRes}! -blur 0.2 -resize 720x$VerticalVideoRes! +matte -depth 8 "$WorkingDir/TransitionFrames/$NextFrameNumber.png"

    FrameCounter=$[$FrameCounter + 1]
    NumFramesLeft=$[ $NumFramesLeft - 1 ]

    echo -n "."
    UpdateProgress $[ ($FrameCounter * 100) / ($NumTransFrames * 2) ]

    done

  DiskSpaceError "$WorkingDir"

  UpdateProgress 100  
  UpdateTask "Encoding to MiniDV Video..."

  pushd "$WorkingDir/TransitionFrames/"

  image2raw * > "$4"

  popd

  UpdateTask "Done."

  DiskSpaceError "$WorkingDir/TransitionFrames/"

  echo "Done".
}


# Video Transition Effect 'Still'
#
# $1 = DV Tail filename
# $2 = DV Head filename
# $3 = Length of Head and Tail parts, in milliseconds
# $4 = Output filename

VideoTransEffect_Still ()

{
  echo -n "Generating Still frames"

  rm -f -R "$WorkingDir/TransitionFrames"
  mkdir "$WorkingDir/TransitionFrames"

  NumTransFrames=$[($3 / $SingleFrameDuration)]

  # Extract last frame of DV tail

  smil2raw -o $[ $NumTransFrames - 1 ] -f 1 "$1" | raw2image -i 0 "$WorkingDir/stillframe.png"

  # Crude de-interlacing
  # We don't want the picture to flicker

  convert "$WorkingDir/stillframe.png" -resize 720x$[$VerticalVideoRes / 2]! -resize 720x$VerticalVideoRes! "$WorkingDir/stillframe.png"

  # Generate a series of still frames

  FrameCounter=0
  NextFrameNumber=`echo $[1000000 + $FrameCounter] | tail -c 7`
  FirstFrameNumber=$NextFrameNumber

  #cp "$WorkingDir/stillframe.png" "$WorkingDir/TransitionFrames/$NextFrameNumber.png"

  FrameCounter=1
  NextFrameNumber=`echo $[1000000 + $FrameCounter] | tail -c 7`
  UpdateTask "Creating Freeze effect..."
  UpdateProgress 0

  for ((Frame=0; Frame < $NumTransFrames ; Frame++))
    do

    ln -s "$WorkingDir/stillframe.png" "$WorkingDir/TransitionFrames/$NextFrameNumber.png"

    FrameCounter=$[$FrameCounter + 1]
    NextFrameNumber=`echo $[1000000 + $FrameCounter] | tail -c 7`

    echo -n "."

    done

  # Write still frames into new dv video file

  UpdateProgress 100  
  UpdateTask "Encoding to MiniDV Video..."

  pushd "$WorkingDir/TransitionFrames/"

  image2raw * > "$WorkingDir/StillVideo.dv"

  popd

  DiskSpaceError "$WorkingDir"

  echo "<?xml version=\"1.0\"?>" > "$WorkingDir/StillSequence.smil"
  echo "<smil xmlns:smil2=\"http://www.w3.org/2001/SMIL20/Language\">" >> "$WorkingDir/StillSequence.smil"
  echo "  <seq>" >> "$WorkingDir/StillSequence.smil"
  echo "    <video src=\"$1\" clipBegin=\"0\" clipEnd=\"$(( $NumTransFrames - 1 ))\"/>" >> "$WorkingDir/StillSequence.smil"
  echo "  </seq>" >> "$WorkingDir/StillSequence.smil"
  echo "  <seq>" >> "$WorkingDir/StillSequence.smil"
  echo "    <video src=\"$WorkingDir/StillVideo.dv\" clipBegin=\"0\" clipEnd=\"$(( $NumTransFrames - 1 ))\"/>" >> "$WorkingDir/StillSequence.smil"
  echo "  </seq>" >> "$WorkingDir/StillSequence.smil"
  echo "</smil> " >> "$WorkingDir/StillSequence.smil"

  smil2raw "$WorkingDir/StillSequence.smil" > "$4"

  UpdateTask "Done."

  DiskSpaceError "$4"

  echo "Done".
}


# Video Transition Effect 'Zoom'
#
# $1 = DV Tail filename
# $2 = DV Head filename
# $3 = Length of Head and Tail parts, in milliseconds
# $4 = Output filename

VideoTransEffect_Zoom ()

{
  echo
  echo "ZOOM: $VerticalVideoRes"
  echo

  echo -n "Generating Zoom frames"

  rm -f -R "$WorkingDir/HeadFrames" "$WorkingDir/TailFrames" "$WorkingDir/TransitionFrames"

  mkdir "$WorkingDir/HeadFrames"
  mkdir "$WorkingDir/TailFrames"
  mkdir "$WorkingDir/TransitionFrames"

  UpdateTask "Extracting frames from video file."

  NumTransFrames=$[($3 / $SingleFrameDuration) * 2]

  pushd "$WorkingDir"

  DiskSpaceWarning "$WorkingDir" $[ $NumTransFrames ]

  cat "$1" | raw2image -i 0 -m TailFrames/.png
  cat "$2" | raw2image -i 0 -m HeadFrames/.png

  popd

  DiskSpaceWarning "$WorkingDir/TransitionFrames/$Frame" $NumTransFrames

  NumFramesLeft=$NumTransFrames

  FrameCounter=1

  convert -size 720x$VerticalVideoRes xc:white "$WorkingDir/White.png"

  FrameNames=( `ls "$WorkingDir/TailFrames"` )
  NumFrames=$(( $NumFramesLeft / 2 ))
  UpdateTask "Create Zoom effect..."
  UpdateProgress 0

  for ((Frame=0; Frame < $NumFrames ; Frame++))
    do

    XZoom=$[ ( ( ( $Frame + 1 ) * 720 ) / $NumFrames ) ]
    YZoom=$[ ( ( ( $Frame + 1 ) * $VerticalVideoRes ) / $NumFrames ) ]

    if [ $XZoom == 720 ]; then
     XZoom=719
    fi

    if [ $YZoom == $VerticalVideoRes ]; then
     YZoom=$[ $VerticalVideoRes - 1 ]
    fi

    NextFrameNumber=$(echo $[1000000 + $FrameCounter] | tail -c 7)

    convert "$WorkingDir/TailFrames/${FrameNames[$Frame]}" -crop $((720 - $XZoom))x$(($VerticalVideoRes - $YZoom))+$[$XZoom / 2]+$[$YZoom / 2] -blur 0.1 -resize 720x$VerticalVideoRes! +matte -depth 8 "$WorkingDir/TransitionFrames/$NextFrameNumber.png"

    FrameCounter=$[$FrameCounter + 1]
    NumFramesLeft=$[ $NumFramesLeft - 1 ]

    echo -n "."
    UpdateProgress $[ ($FrameCounter * 100) / ($NumTransFrames * 2) ]

    done

  DiskSpaceError "$WorkingDir"

  FrameNames=( `ls "$WorkingDir/HeadFrames"` )
  NumFrames=$NumFramesLeft

  for ((Frame=0; Frame < $NumFrames ; Frame++))
    do
    XZoom=$(( 720 - ((($Frame + 1) * 720) / $NumFrames ) ))
    YZoom=$(( $VerticalVideoRes - ((($Frame + 1) * $VerticalVideoRes) / $NumFrames ) ))

    if [ $XZoom == 720 ]; then
     XZoom=719
    fi
    if [ $YZoom == $VerticalVideoRes ]; then
     YZoom=$[ $VerticalVideoRes + 1 ]
    fi

    NextFrameNumber=`echo $[1000000 + $FrameCounter] | tail -c 7`

    convert "$WorkingDir/HeadFrames/${FrameNames[$Frame]}" -crop $((720 - $XZoom))x$(($VerticalVideoRes - $YZoom))+$[$XZoom / 2]+$[$YZoom / 2] -blur 0.1 -resize 720x$VerticalVideoRes! +matte -depth 8 "$WorkingDir/TransitionFrames/$NextFrameNumber.png"

    FrameCounter=$[$FrameCounter + 1]
    NumFramesLeft=$[ $NumFramesLeft - 1 ]

    echo -n "."
    UpdateProgress $[ ($FrameCounter * 100) / ($NumTransFrames * 2) ]

    done

  DiskSpaceError "$WorkingDir"
  UpdateProgress 100  
  UpdateTask "Encoding to MiniDV Video..."

  pushd "$WorkingDir/TransitionFrames/"

  image2raw * > "$4"

  popd

  UpdateTask "Done."

  DiskSpaceError "$WorkingDir/TransitionFrames/"

  echo "Done".
}


# Audio Transition Effect 'Crossfade'
#
# $1 = PCM Wave Tail filename
# $2 = PCM Wave Head filename
# $3 = Length of Head and Tail parts, in milliseconds
# $4 = Output filename

AudioTransEffect_Crossfade ()

{
  # sox fade-out does not work when the fade length equals the cut-off length.
  # Also seems to work better when times are not given in samples but seconds.

  echo "Creating crossfade soundtrack..."
  UpdateTask "Creating audio transition..."

  sox "$1" -t .wav "$WorkingDir/FadedTail.wav" fade t 0 $(( 48 * $3 ))s $(( 48 * ( $3 ) ))s
  sox "$2" -t .wav "$WorkingDir/FadedHead.wav" fade t $(( 48 * $3 ))s

  # Watch it: Sox does not add the two, but it takes average samples. Need to
  # multiply volume to keep the sound transition smooth.

  soxmix -v 1.0 "$WorkingDir/FadedTail.wav" -v 1.0 "$WorkingDir/FadedHead.wav" -r 48000 -c 2 -s -w "$4"

  UpdateTask "Done."

  DiskSpaceError "$WorkingDir"

  rm -f "$WorkingDir/FadedTail.wav" "$WorkingDir/FadedHead.wav"

  echo "Done".
}

# Audio Transition Effect 'FadeoutFadein'
#
# $1 = PCM Wave Tail filename
# $2 = PCM Wave Head filename
# $3 = Length of Head and Tail parts, in milliseconds
# $4 = Output filename

AudioTransEffect_FadeoutFadein ()

{
  # sox fade-out does not work when the fade length equals the cut-off length.
  # Also seems to work better when times are not given in samples but seconds.

  echo "Creating fade-out / fade-in soundtrack..."
  UpdateTask "Creating sound transition..."

  sox "$1" -r 48000 -c 2 -s -w "$WorkingDir/FadedTail.raw" fade t 0 $(( 48 * $3 ))s $(( 48 * ( $3 - 1 ) ))s
  sox "$2" -r 48000 -c 2 -s -w "$WorkingDir/FadedHead.raw" fade t $(( 48 * $3 ))s

  cat "$WorkingDir/FadedTail.raw" "$WorkingDir/FadedHead.raw" > "$4"

  UpdateTask "Done."

  DiskSpaceError "$WorkingDir"

  rm -f "$WorkingDir/FadedHead.raw" "$WorkingDir/FadedTail.raw"
  echo "Done".
}


CreateAudioTransEffect ()

{
 if [ $1 == 7 ]; then

   # Video effect is Flash

   AudioTransEffect_FadeoutFadein "$2" "$3" $4 "$5"
 
 elif [ $1 == 2 ]; then

   # Video effect is Crossfade

   AudioTransEffect_Crossfade "$2" "$3" $4 "$5"

 elif [ $1 == 3 ]; then

   # Video effect is PushAside

   AudioTransEffect_Crossfade "$2" "$3" $4 "$5"

 elif [ $1 == 4 ]; then

   # Video effect is Blur

   AudioTransEffect_FadeoutFadein "$2" "$3" $4 "$5"

 elif [ $1 == 5 ]; then

   # Video effect is Zoom

   AudioTransEffect_FadeoutFadein "$2" "$3" $4 "$5"
 
 elif [ $1 == 6 ]; then

   # Video effect is Still

   AudioTransEffect_FadeoutFadein "$2" "$3" $4 "$5"
 
 elif [ $1 == 1 ]; then

   # No video effect

   return
 
 fi

}

CreateVideoTransEffect ()

{
 if [ $1 == 7 ]; then

   # Effect is Flash

   VideoTransEffect_Flash "$2" "$3" $4 "$5"
 
 elif [ $1 == 2 ]; then

   # Effect is Crossfade

   VideoTransEffect_Crossfade "$2" "$3" $4 "$5"

 elif [ $1 == 3 ]; then

   # Effect is PushAside

   VideoTransEffect_PushAside "$2" "$3" $4 "$5"

 elif [ $1 == 4 ]; then

   # Effect is Blur

   VideoTransEffect_Blur "$2" "$3" $4 "$5"

 elif [ $1 == 5 ]; then

   # Effect is Zoom

   VideoTransEffect_Zoom "$2" "$3" $4 "$5"
 
 elif [ $1 == 6 ]; then

   # Effect is Still

   VideoTransEffect_Still "$2" "$3" $4 "$5"
 
 elif [ $1 == 1 ]; then
 
   # No effect

   return
 
 fi

}


CalcHeadTailLengths ()
{
  for ((Transition=1; Transition <= NumTransitions ; Transition++))
    do

    if [ ${TransitionEffects[$Transition]} == 2 ]; then 

      # Crossfade

      HeadTailLengths[$Transition]=${TransitionLengths[$Transition]}

    elif [ ${TransitionEffects[$Transition]} == 3 ]; then 

      # PushAside

      HeadTailLengths[$Transition]=${TransitionLengths[$Transition]}

    elif [ ${TransitionEffects[$Transition]} == 7 ]; then 

      # Flash

      HeadTailLengths[$Transition]=$(( ${TransitionLengths[$Transition]} / 2 ))

    elif [ ${TransitionEffects[$Transition]} == 4 ]; then 

      # Blur

      HeadTailLengths[$Transition]=$(( ${TransitionLengths[$Transition]} / 2 ))

    elif [ ${TransitionEffects[$Transition]} == 5 ]; then 

      # Zoom

      HeadTailLengths[$Transition]=$(( ${TransitionLengths[$Transition]} / 2 ))

    elif [ ${TransitionEffects[$Transition]} == 6 ]; then 

      # Still

      HeadTailLengths[$Transition]=$(( ${TransitionLengths[$Transition]} / 2 ))

    elif [ ${TransitionEffects[$Transition]} == 1 ]; then   

      # No effect

      HeadTailLengths[$Transition]=0
      TransitionLengths[$Transition]=0

    fi

    echo "Transition $Transition will have length of ${TransitionLengths[$Transition]}."
    echo "${HeadTailLengths[$Transition]} ms of head and tail material required."

    done
}

CreateTransitions ()

{

  for ((Transition=1; Transition <= NumTransitions ; Transition++))
    do

    # Extract the head and tail of the video files

    echo "Transition $Transition: Extracting video head and tail..."

    DVTail "${DVFiles[$Transition]}" ${HeadTailLengths[$Transition]} "$WorkingDir/Tail.dv"
    DVHead "${DVFiles[$[$Transition + 1]]}" ${HeadTailLengths[$Transition]} "$WorkingDir/Head.dv"

    echo "Transition $Transition: Extracting audio head and tail..."

    SoundTail "${DVFiles[$Transition]}" ${HeadTailLengths[$Transition]} "$WorkingDir/Tail.wav"
    SoundHead "${DVFiles[$[$Transition + 1]]}" ${HeadTailLengths[$Transition]} "$WorkingDir/Head.wav"

    # Create transitions

    echo "Creating video transition $Transition:"

    CreateVideoTransEffect ${TransitionEffects[$Transition]} \
                           "$WorkingDir/Tail.dv" "$WorkingDir/Head.dv" \
			   ${HeadTailLengths[$Transition]} "$WorkingDir/Transition$Transition.dv"

    echo "Creating audio transition $Transition:"

    CreateAudioTransEffect ${TransitionEffects[$Transition]} \
                           "$WorkingDir/Tail.wav" "$WorkingDir/Head.wav" \
			   ${HeadTailLengths[$Transition]} "$WorkingDir/Transition$Transition.raw"

    rm -f "$WorkingDir/Tail.dv" "$WorkingDir/Head.dv" \
          "$WorkingDir/Tail.wav" "$WorkingDir/Head.wav"

  done

  echo "Completed creating all transition effects."
}

# Concatenate all sound parts into one BIG file.

AssembleAudio ()

{
  echo "Extracting audio fragments"
  UpdateTask "Gathering Audio Fragments..."

  RenderCommand=""

  # First, we extract all sound from the DV video's that is *NOT* involved in any
  # transition effect.

  NumBytes=$[ `smil2wav 2>/dev/null "${DVFiles[1]}" | wc -c` - $SMIL2WavHeaderSize ]

  DetectSoundProperties "${DVFiles[1]}"

  NumBodyBytes=$[$NumBytes - ( ($DetectedBytesPerSecond * ${HeadTailLengths[1]}) / 1000 ) ]

  echo "AUDIO FRAGMENT 1: $NumBytes Bytes, `echo "scale=1; $NumBytes/192" | bc` milliseconds."
  echo "            BODY: $NumBodyBytes Bytes, `echo "scale=1; $NumBodyBytes/192" | bc` milliseconds."
  echo "        STRIPPED: $((( ($DetectedBytesPerSecond * ${HeadTailLengths[1]}) / 1 ))) Bytes, `echo "scale=2; $((( ($DetectedBytesPerSecond * ${HeadTailLengths[1]}) / 1000 )))/192" | bc` milliseconds."

  if [ $NumBodyBytes -gt 0 ]; then

    if [ ! $DetectedSampleRate == 48000 ]; then

      echo "Non-standard sampling rate detected, resampling to 48 KHz..."

      RenderCommand="$RenderCommand smil2wav \"${DVFiles[1]}\" 2>/dev/null | dd bs=$SMIL2WavHeaderSize skip=1 | head -c $NumBodyBytes | sox -r $DetectedSampleRate -s -c 2 -w -t .raw /dev/stdin -r 48000 -s -c 2 -w -t .raw /dev/stdout resample ;"

    else

      RenderCommand="$RenderCommand smil2wav \"${DVFiles[1]}\" 2>/dev/null | dd bs=$SMIL2WavHeaderSize skip=1 | head -c $NumBodyBytes | sox -r $DetectedSampleRate -s -c 2 -w -t .raw /dev/stdin -r 48000 -s -c 2 -w -t .raw /dev/stdout ;"

    fi

  fi

  # Add sound of first transition, if there is one.

  if [ -e "$WorkingDir/Transition1.raw" ]; then
    RenderCommand="$RenderCommand cat \"$WorkingDir/Transition1.raw\" ;"
  fi

  # Now add the body/transition sound fragments of all other files

  for ((File=2; File <= $[ $NumDVFiles - 1] ; File++))
    do

      NumBytes=$[ `smil2wav 2>/dev/null "${DVFiles[$File]}" | wc -c` - $SMIL2WavHeaderSize ]

      DetectSoundProperties "${DVFiles[$File]}"

      NumBodyBytes=$[$NumBytes - ( ( ( $DetectedBytesPerSecond * ${HeadTailLengths[$(($File - 1))]} + $DetectedBytesPerSecond * ${HeadTailLengths[$File]} ) ) / 1000 ) ]

      echo "AUDIO FRAGMENT $File: $NumBytes Bytes, `echo "scale=1; $NumBytes/192" | bc` milliseconds."
      echo "            BODY: $NumBodyBytes Bytes, `echo "scale=1; $NumBodyBytes/192" | bc` milliseconds."
      echo "        STRIPPED: $((( ( ( $DetectedBytesPerSecond * ${HeadTailLengths[$(($File - 1))]} + $DetectedBytesPerSecond * ${HeadTailLengths[$File]} ) ) / 1000 ))) Bytes, `echo "scale=2; $((( ( ( $DetectedBytesPerSecond * ${HeadTailLengths[$(($File - 1))]} + $DetectedBytesPerSecond * ${HeadTailLengths[$File]} ) ) / 1000 )))/192" | bc` milliseconds."

      if [ $NumBodyBytes -gt 0 ]; then

        if [ ! $DetectedSampleRate == 48000 ]; then

          echo "Non-standard sampling rate detected, resampling to 48 KHz..."

          RenderCommand="$RenderCommand smil2wav \"${DVFiles[$File]}\" 2>/dev/null | dd bs=$SMIL2WavHeaderSize skip=1 | tail -c $[$NumBytes - ( ($DetectedBytesPerSecond * ${HeadTailLengths[$(($File - 1))]}) / 1000 ) ] | head -c $NumBodyBytes | sox -r $DetectedSampleRate -s -c 2 -w -t .raw /dev/stdin -r 48000 -s -c 2 -w -t .raw /dev/stdout resample ;"

        else

          RenderCommand="$RenderCommand smil2wav \"${DVFiles[$File]}\" 2>/dev/null | dd bs=$SMIL2WavHeaderSize skip=1 | tail -c $[$NumBytes - ( ($DetectedBytesPerSecond * ${HeadTailLengths[$(($File - 1))]}) / 1000 ) ] | head -c $NumBodyBytes | sox -r $DetectedSampleRate -s -c 2 -w -t .raw /dev/stdin -s -c 2 -w -t .raw -r 48000 /dev/stdout ;"

        fi
      fi

      # Add transition sound:

      if [ -e "$WorkingDir/Transition$File.raw" ]; then
        RenderCommand="$RenderCommand cat \"$WorkingDir/Transition$File.raw\" ;"
      fi

    done

  # Add sound of last body/transition pair 

  NumBytes=$[ `smil2wav "${DVFiles[$NumDVFiles]}" 2>/dev/null | wc -c` - $SMIL2WavHeaderSize ]

  DetectSoundProperties "${DVFiles[$NumDVFiles]}"

  NumBodyBytes=$[$NumBytes - ( ($DetectedBytesPerSecond * ${HeadTailLengths[$NumTransitions]}) / 1000 ) ]

  echo "AUDIO FRAGMENT $NumDVFiles: $NumBytes Bytes, `echo "scale=1; $NumBytes/192" | bc` milliseconds."
  echo "            BODY: $NumBodyBytes Bytes, `echo "scale=1; $NumBodyBytes/192" | bc` milliseconds."

  if [ $NumBodyBytes -gt 0 ]; then

    if [ ! $DetectedSampleRate == 48000 ]; then

      echo "Non-standard sampling rate detected, resampling to 48 KHz..."

      RenderCommand="$RenderCommand smil2wav \"${DVFiles[$NumDVFiles]}\" 2>/dev/null | dd bs=$SMIL2WavHeaderSize skip=1 | tail -c $NumBodyBytes | sox -r $DetectedSampleRate -s -c 2 -w -t .raw /dev/stdin -r 48000 -s -c 2 -w -t .raw /dev/stdout resample ;"

    else

      RenderCommand="$RenderCommand smil2wav \"${DVFiles[$NumDVFiles]}\" 2>/dev/null | dd bs=$SMIL2WavHeaderSize skip=1 | tail -c $NumBodyBytes | sox -r $DetectedSampleRate -s -c 2 -w -t .raw /dev/stdin -r 48000 -s -c 2 -w -t .raw /dev/stdout ;"

    fi

  fi

  # Place entire command inside curly brackets

  RenderCommand="( $RenderCommand )"

  echo ""
  echo "RenderCommand is $RenderCommand."
  echo ""

  # Execute command and encode to wav format the output using sox:

  eval "$RenderCommand" | sox -r 48000 -c 2 -s -w -t .raw /dev/stdin "$WorkingDir/Final.wav"

}

AssembleVideo ()

{
  echo "Assembling final audio video file"

  ChapterString="0:0:0"
  FrameCounter=0

  NumFrames=$[ `du -b "${DVFiles[1]}" | awk '{print $1}'` / $DVFrameSize ]

  # Write a Kino SMIL project file:

  echo "<?xml version=\"1.0\"?>" > "$WorkingDir/Final.smil"
  echo "<smil xmlns:smil2=\"http://www.w3.org/2001/SMIL20/Language\">" >> "$WorkingDir/Final.smil"

  ClipEnd=$(($NumFrames - (${HeadTailLengths[1]} / $SingleFrameDuration) - 1 ))

  # Write first video sequence:

  if [ $ClipEnd -ge 0 ]; then

    echo "  <seq>" >> "$WorkingDir/Final.smil"
    echo "    <video src=\"${DVFiles[1]}\" clipBegin=\"0\" clipEnd=\"$ClipEnd\"/>" >> "$WorkingDir/Final.smil"
    echo "  </seq>" >> "$WorkingDir/Final.smil"

    FrameCounter=$[$FrameCounter + $ClipEnd]

  fi

  # Now add all transition/body sequences

  for ((File=2; File <= $(($NumDVFiles - 1)); File++))
    do

      NumFrames=$[ `du -b "${DVFiles[$File]}" | awk '{print $1}'` / $DVFrameSize ]

      ClipEnd=$[ (${TransitionLengths[$(($File - 1))]} / $SingleFrameDuration) - 1 ]

      if [ $ClipEnd -ge 0 ]; then

        echo "  <seq>" >> "$WorkingDir/Final.smil"
        echo "    <video src=\"$WorkingDir/Transition$(($File - 1)).dv\" clipBegin=\"0\" clipEnd=\"$ClipEnd\"/>" >> "$WorkingDir/Final.smil"
        echo "  </seq>" >> "$WorkingDir/Final.smil"

        # Add chapter marker at the beginning of the transition
        ChapterString="$ChapterString,$(TimeIndex $FrameCounter)"

        FrameCounter=$[$FrameCounter + $ClipEnd]
      fi

      ClipBegin=$[ ${HeadTailLengths[$[$File - 1]]} / $SingleFrameDuration ]
      ClipEnd=$[ $NumFrames - (${HeadTailLengths[$File]} / $SingleFrameDuration) - 1 ]

      if [ $ClipEnd -ge 0 ]; then
      
        echo "  <seq>" >> "$WorkingDir/Final.smil"
        echo "    <video src=\"${DVFiles[$File]}\" clipBegin=\"$ClipBegin\" clipEnd=\"$ClipEnd\"/>" >> "$WorkingDir/Final.smil"
        echo "  </seq>" >> "$WorkingDir/Final.smil"

        FrameCounter=$[$FrameCounter + $ClipEnd - $ClipBegin]

      fi
  done

  # Add last transition:

  NumFrames=$[ `du -b "${DVFiles[$NumDVFiles]}" | awk '{print $1}'` / $DVFrameSize ]

  ClipEnd=$(( (${TransitionLengths[$NumTransitions]} / $SingleFrameDuration) - 1))

  if [ $ClipEnd -ge 0 ]; then

    echo "  <seq>" >> "$WorkingDir/Final.smil"
    echo "    <video src=\"$WorkingDir/Transition$NumTransitions.dv\" clipBegin=\"0\" clipEnd=\"$ClipEnd\"/>" >> "$WorkingDir/Final.smil"
    echo "  </seq>" >> "$WorkingDir/Final.smil"

    # Add chapter marker at the beginning of the transition
    ChapterString="$ChapterString,$(TimeIndex $FrameCounter)"

    FrameCounter=$[$FrameCounter + $ClipEnd]

  fi

  ClipBegin=$[ ${HeadTailLengths[$NumTransitions]} / $SingleFrameDuration ]
  ClipEnd=$[ $NumFrames - 1 ]

  if [ $ClipEnd -ge 0 ]; then

    echo "  <seq>" >> "$WorkingDir/Final.smil"
    echo "    <video src=\"${DVFiles[$NumDVFiles]}\" clipBegin=\"$ClipBegin\" clipEnd=\"$ClipEnd\"/>" >> "$WorkingDir/Final.smil"
    echo "  </seq>" >> "$WorkingDir/Final.smil"

    FrameCounter=$[$FrameCounter + $ClipEnd - $ClipBegin]
  fi

  echo "</smil> " >> "$WorkingDir/Final.smil"

  # Assemble the final video in one go:

  echo "Rendering..."
  UpdateTask "Merging / encoding video fragments..."

  FullOutputFileName="$OutputDir/$FinalVideoName.dv"
  OutputFileName="`basename "$FullOutputFileName"`"

  cd "$OutputDir"

  smil2raw "$WorkingDir/Final.smil" | dubdv "$WorkingDir/Final.wav" > "$OutputFileName"

  DiskSpaceError "$OutputFileName"

  # Write a Chapter description file for KmPg2 MPEG2 encoder:

  ChapterFile="${OutputFileName%%.*}.chapters"

  echo
  echo "Writing chapter definition file: $ChapterFile"
  echo
  echo "$ChapterString"
  echo

  echo "$ChapterString" > "$ChapterFile"

  rm -f "$WorkingDir/Final.wav"

  echo "Done."
  UpdateTask "Done."

}

# Check if video files are compatible ( no mixing
# of 16:9 and 4:3 or PAL and NTSC). Also store some
# video properties in global variables.

CheckVideoProperties ()
{
  UpdateTask "Detecting video properties..."

  DetectVideoRes "${DVFiles[1]}"
  VideoType=$(DetectVideoType)
  VideoAspect=$(DetectVideoAspect "${DVFiles[1]}")

  if [ -z "$VideoType" ]; then

    PrintError "Failed to determine video type for file ${DVFiles[1]}."

  fi

  echo "First video file is of type $VideoType and aspect ratio $VideoAspect."

  for ((File=2; File <= $(($NumDVFiles )); File++))
    do
      echo "Video $File is of type $VideoType"

      DetectVideoRes "${DVFiles[$File]}"

      if [ "$(DetectVideoType)" != $VideoType ]; then

        PrintError "The video files are not of the same type.\n\nMake sure that all video files are either PAL or NTSC video. Mixing PAL and NTSC material is not possible."

      fi

      if [ "$(DetectVideoAspect "${DVFiles[$File]}")" != $VideoAspect ]; then

        PrintError "The video files do not have the same display aspect ratio.\n\nMake sure that all video files are either in widescreen format or standard 4:3. Mixing widescreen recordings with normal recordings is not possible."

      fi

    done

 # DV frames have fixed size, exactly 120.000 bytes 
 # per frame for NTSC and 144.000 bytes per frame for PAL.

 if [ $VideoType == "PAL" ]; then

  DVFrameSize=144000
  SingleFrameDuration=40
  FrameRate="25.0"

 else

  DVFrameSize=120000
  SingleFrameDuration=33
  FrameRate="29.97"

 fi

 UpdateTask "Done."
}

# Checks for overlapping transition effects

SanityCheck ()

{
  echo -n "Checking sanity of user input..."
  UpdateTask "Performing sanity checks..."

  NumFrames=$[`du -b "${DVFiles[1]}" | awk '{print $1}'` / $DVFrameSize ]
  
  echo "Number of frames found to be $NumFrames."
  echo "HeadTailLengths[1] is ${HeadTailLengths[1]}."
  
  if [ $[${HeadTailLengths[1]} / $SingleFrameDuration] -gt $NumFrames ]; then
  
   HeadTailLengths[1]=0
   TransitionEffects[1]=1

   PrintWarning "The first transition is extending beyond the start of the first video fragment. No transition effect will be created for the first transition.\n\nContinue?"

  fi

  for ((File=2; File <= $(($NumDVFiles - 1)); File++))
    do

      NumFrames=$[`du -b "${DVFiles[$File]}" | awk '{print $1}'` / $DVFrameSize ]

      if [ $[ (${HeadTailLengths[$(($File - 1))]} / $SingleFrameDuration) + (${HeadTailLengths[$File]} / $SingleFrameDuration)] -gt $NumFrames ]; then

         # Reset transition types to "None"

         TransitionEffects[$(($File - 1))]=1
         HeadTailLengths[$(($File - 1))]=0

	 TransitionEffects[$File]=1
	 HeadTailLengths[$File]=0

         PrintWarning "Transition effects $(($File - 1)) and $File are overlapping. None of these transition effects will be created.\n\nContinue?"

      fi

    done

  NumFrames=$[ `du -b "${DVFiles[$NumDVFiles]}" | awk '{print $1}'` / $DVFrameSize ]

  if [ $[${HeadTailLengths[$[$NumDVFiles - 1]]} / $SingleFrameDuration] -gt $NumFrames ]; then

   TransitionEffects[$[$NumDVFiles - 1]]=1
   HeadTailLengths[$[$NumDVFiles - 1]]=0

   PrintWarning "The last transition effect is extending beyond the end of the last video fragment. This transition effect will not be created.\n\nContinue?"

  fi

 echo "Input OK" 

 UpdateTask "Done."
}

# Transition lengts must be multiples of two frames.

FixTransitionLengths ()

{
  DoubleFrameDuration=$[$SingleFrameDuration * 2]

  for ((Transition=1; Transition <= NumTransitions ; Transition++))
    do

    if [ $[${TransitionLengths[$Transition]} % $DoubleFrameDuration] -gt 0 ]; then

      TransitionLengths[$Transition]=$[${TransitionLengths[$Transition]} - ( ${TransitionLengths[$Transition]} % $DoubleFrameDuration ) ]

      echo "Corrected transition length $Transition to ${TransitionLengths[$Transition]}"

    fi

    done

}

ReadConfigFile ()

{
  if [ -z "$1" ]; then
   echo "You did not provide a configuration file."
   echo
   echo "USAGE: SlideShow ConfigFile.conf"
   exit -1
  fi

  NumDVFiles=$(cat "$1" | grep -E ^NumDVFiles | cut -d = -f 2)
  NumTransitions=$[$NumDVFiles - 1]
  OutputDir="$( cat "$1" | grep -E ^OutputDir= | cut -d \" -f 2)"
  FinalVideoName="$( cat "$1" | grep -E ^FinalVideoName= | cut -d \" -f 2)"

  for ((File=1; File <= NumDVFiles ; File++))
   do
   DVFiles[$File]="$( cat "$1" | grep -F DVFiles[$File]  | cut -d \" -f 2)"
   echo "DV File $File: ${DVFiles[$File]}"
   done

  for ((Transition=1; Transition <= NumTransitions ; Transition++))
   do
   TransitionEffects[$Transition]="$( cat "$1" | grep -F TransitionEffects[$Transition]  | cut -d \" -f 2)"
   echo "Transition Effect $Transition: ${TransitionEffects[$Transition]}"
   TransitionLengths[$Transition]="$( cat "$1" | grep -F TransitionLengths[$Transition]  | cut -d \" -f 2)"
   echo "Transition Length $Transition: ${TransitionLengths[$Transition]}"
   done

} 

NumRandomEffects=9

# Array of 4 Crossfade effects, 2 Blur effects, 
# one PushAside, one Zoom and one Flash.
# Used to pick effects while some effects are picked
# more often than others.

RandomEffectList[0]=2
RandomEffectList[1]=2
RandomEffectList[2]=2
RandomEffectList[3]=2
RandomEffectList[4]=4
RandomEffectList[5]=4
RandomEffectList[6]=3
RandomEffectList[7]=5
RandomEffectList[8]=7

# If a transition effect is 'Random', replace it
# with a random effect.

ChooseRandomEffects ()

{
  for ((Transition=1; Transition <= NumTransitions ; Transition++))
    do

    if [ ${TransitionEffects[$Transition]} == 0 ]; then

      # User selected 'Random' effect

      RandomEffect=${RandomEffectList[$[( ( $RANDOM * ($NumRandomEffects - 1) ) / 32767)]]}
      echo "Randomly selected effect $RandomEffect for transition $Transition"
      TransitionEffects[$Transition]=$RandomEffect
    fi
    done
}

# Script starts here:

DialogDCOPProgressCommand="$2"
DialogDCOPTaskCommand="$3"

RunChecks
ReadConfigFile "$1"
CheckVideoProperties
ChooseRandomEffects
FixTransitionLengths
CalcHeadTailLengths
SanityCheck


WorkingDir=`echo "$OutputDir/DVAssembler_Temporary_Files_$RANDOM"`
rm -f -R "$WorkingDir"
mkdir -p "$WorkingDir"

CreateTransitions
AssembleAudio
AssembleVideo

#rm -f -R "$WorkingDir"

echo "Assembly complete. Have a nice day!"
UpdateTask "Assembly Complete."

